from edsl.trackers.Tracker import Tracker
from edsl.utilities.interface import heartbeat_generator, human_readable_labeler_creator


class TrackerTasks(Tracker):
    """ """

    class WorkerActivated(Tracker.TrackerEvent):
        def apply(self, tracker):
            with tracker.lock:
                tracker.active_workers += 1

    class WorkerDeactivated(Tracker.TrackerEvent):
        def apply(self, tracker):
            with tracker.lock:
                tracker.active_workers -= 1

    class TaskStarted(Tracker.TrackerEvent):
        def apply(self, tracker):
            with tracker.lock:
                tracker.interviews_started += 1

    class TaskCompleted(Tracker.TrackerEvent):
        def apply(self, tracker):
            with tracker.lock:
                tracker.interviews_completed += 1

    class ThreadStatus(Tracker.TrackerEvent):
        def __init__(self, thread_id, status):
            self.thread_id = thread_id
            self.status = status

        def apply(self, tracker):
            with tracker.lock:
                tracker.thread_status[self.thread_id] = self.status

    def __init__(self, num_interviews, lock, monitored_queue):
        self.num_interviews = num_interviews  # how many interviews are there in total
        self.lock = lock
        self.hearbeat = heartbeat_generator()  # generator to show responses coming in
        self.active_workers = 0
        self.interviews_started = 0
        self.interviews_completed = 0
        self.thread_status = dict()
        # this is a function that takes thread IDs and turns them into successive integers
        self.human_readable_labeler = human_readable_labeler_creator()

        super().__init__(monitored_queue)

    def tracked_values(self) -> dict:
        return {
            "Active Workers": self.active_workers,
            "Interviews Started": self.interviews_started,
            "Completed Interviews": self.interviews_completed,
        }

    def allowed_events(self):
        return set(
            [
                TrackerTasks.WorkerActivated,
                TrackerTasks.WorkerDeactivated,
                TrackerTasks.TaskStarted,
                TrackerTasks.TaskCompleted,
                TrackerTasks.ThreadStatus,
            ]
        )

    @property
    def percentage_complete(self):
        return round(100 * self.interviews_completed / self.num_interviews, 2)



from docx import Document


class SurveyExportMixin:
    def docx(self) -> "Document":
        "Generates a docx document for the survey"
        doc = Document()
        doc.add_heading("EDSL Auto-Generated Survey")
        doc.add_paragraph(f"\n")
        for index, question in enumerate(self._questions):
            h = doc.add_paragraph()  # Add question as a paragraph
            h.add_run(f"Question {index + 1} ({question.question_name})").bold = True
            h.add_run(f"; {question.question_type}").italic = True
            p = doc.add_paragraph()
            p.add_run(question.question_text)
            if question.question_type == "linear_scale":
                for key, value in getattr(question, "option_labels", {}).items():
                    doc.add_paragraph(str(key) + ": " + str(value), style="ListBullet")
            else:
                if hasattr(question, "question_options"):
                    for option in getattr(question, "question_options", []):
                        doc.add_paragraph(str(option), style="ListBullet")
        return doc

    def code(self) -> list[str]:
        ## TODO: Refactor to only use the questions actually in the survey
        "Creates the Python code representation of a survey"
        header_lines = ["from edsl.surveys.Survey import Survey"]
        header_lines.append(
            "from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice"
        )
        header_lines.append(
            "from edsl.questions.QuestionFreeText import QuestionFreeText"
        )
        header_lines.append(
            "from edsl.questions.derived.QuestionLinearScale import QuestionLinearScale"
        )
        header_lines.append(
            "from edsl.questions.QuestionNumerical import QuestionNumerical"
        )
        header_lines.append(
            "from edsl.questions.QuestionCheckBox import QuestionCheckBox"
        )
        header_lines.append(
            "from edsl.questions.derived.QuestionYesNo import QuestionYesNo"
        )
        lines = ["\n".join(header_lines)]
        for question in self._questions:
            lines.append(f"{question.question_name} = " + repr(question))
        lines.append(f"survey = Survey(questions = [{', '.join(self.question_names)}])")
        return lines

    def html(self) -> str:
        "Generates the html for the survey"
        html_text = []
        for question in self._questions:
            html_text.append(
                f"<p><b>{question.question_name}</b> ({question.question_type}): {question.question_text}</p>"
            )
            html_text.append("<ul>")
            for option in getattr(question, "question_options", []):
                html_text.append(f"<li>{option}</li>")
            html_text.append("</ul>")
        return "\n".join(html_text)



from edsl.report.ReportOutputs import RegisterElementMeta as registery


class ResultsOutputMixin:
    def add_output_functions(self) -> None:
        """Iterates through all output classes and adds a function to the Results class for each one."""
        output_classes = registery.get_registered_classes().values()
        self.analysis_options = []
        for output_class in output_classes:
            new_function_name = output_class.function_name
            new_function = output_class.create_external_function(self)
            self.__dict__[new_function_name] = new_function

            self.analysis_options.append({new_function_name: output_class.__doc__})



from __future__ import annotations
import random
import textwrap
from jinja2 import Template
from typing import Any, Optional, Union
from edsl.questions import Question
from edsl.exceptions import QuestionAnswerValidationError
from edsl.scenarios import Scenario
from edsl.utilities.utilities import random_string
from edsl.questions.descriptors import (
    QuestionOptionsDescriptor,
    NumSelectionsDescriptor,
)


class QuestionRank(Question):
    """
    This question asks the user to rank options from a list.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_options` are the options the user should select from (list of strings)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `num_selections` is the number of options that must be selected (positive integer)
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionRank.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, run `QuestionRank.example()`
    """

    question_type = "rank"
    question_options: list[str] = QuestionOptionsDescriptor()
    num_selections = NumSelectionsDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
        num_selections: Optional[int] = None,
        short_names_dict: Optional[dict[str, str]] = None,
    ):
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options
        self.short_names_dict = short_names_dict or dict()
        self.num_selections = num_selections or len(question_options)

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: Any) -> dict[str, list[int]]:
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value(answer, "answer", list)
        self.validate_answer_rank(answer)
        return answer

    def translate_answer_code_to_answer(
        self, answer_codes, scenario: Scenario = None
    ) -> list[str]:
        """Translates the answer code to the actual answer."""
        scenario = scenario or Scenario()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]
        translated_codes = []
        for answer_code in answer_codes:
            translated_codes.append(translated_options[int(answer_code)])
        return translated_codes

    def simulate_answer(self, human_readable=True) -> dict[str, Union[int, str]]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            selected = random.sample(self.question_options, self.num_selections)
        else:
            selected = random.sample(
                range(len(self.question_options)), self.num_selections
            )
        answer = {
            "answer": selected,
            "comment": random_string(),
        }
        return answer

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionRank:
        return cls(
            question_name="rank_foods",
            question_text="Rank your favorite foods.",
            question_options=["Pizza", "Pasta", "Salad", "Soup"],
            num_selections=2,
        )


def main():
    from edsl.questions.QuestionRank import QuestionRank

    q = QuestionRank.example()
    q.question_text
    q.question_name
    q.question_options
    q.num_selections
    # validate an answer
    answer = {"answer": [0, 1], "comment": "I like pizza and pasta."}
    q.validate_answer(answer)
    # translate an answer code to an answer
    q.translate_answer_code_to_answer([0, 1])
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from abc import ABC, abstractmethod
import re
from typing import Any, Callable
from edsl.exceptions import (
    QuestionCreationValidationError,
    QuestionAnswerValidationError,
)
from edsl.questions.settings import Settings
from edsl.utilities.utilities import is_valid_variable_name

from edsl.prompts import get_classes

################################
# Helper functions
################################


def contains_single_braced_substring(s: str) -> bool:
    """Checks if the string contains a substring in single braces."""
    pattern = r"(?<!\{)\{[^{}]+\}(?!\})"
    match = re.search(pattern, s)
    return bool(match)


def is_number(value: Any) -> bool:
    """Checks if an object is a number."""
    return isinstance(value, int) or isinstance(value, float)


def is_number_or_none(value: Any) -> bool:
    """Checks if an object is a number or None."""
    return value is None or is_number(value)


################################
# Descriptor ABC
################################


class BaseDescriptor(ABC):
    """ABC for something."""

    @abstractmethod
    def validate(self, value: Any) -> None:
        """Validates the value. If it is invalid, raises an exception. If it is valid, does nothing."""
        pass

    def __get__(self, instance, owner):
        """"""
        if self.name not in instance.__dict__:
            return {}
        return instance.__dict__[self.name]

    def __set__(self, instance, value: Any) -> None:
        self.validate(value, instance)
        instance.__dict__[self.name] = value
        if self.name == "_instructions":
            instructions = value
            if value is not None:
                instance.__dict__[self.name] = instructions
                instance.set_instructions = True
            else:
                potential_prompt_classes = get_classes(
                    question_type=instance.question_type
                )
                if len(potential_prompt_classes) > 0:
                    instructions = potential_prompt_classes[0]().text
                    instance.__dict__[self.name] = instructions
                    instance.set_instructions = False
                else:
                    if not hasattr(instance, "default_instructions"):
                        raise Exception(
                            "No default instructions found and no matching prompts!"
                        )
                    instructions = instance.default_instructions
                    instance.__dict__[self.name] = instructions
                    instance.set_instructions = False

            # instance.set_instructions = value != instance.default_instructions

    def __set_name__(self, owner, name: str) -> None:
        self.name = "_" + name


################################
# General descriptors
################################


class FunctionDescriptor(BaseDescriptor):
    def validate(self, value: Any, instance) -> Callable:
        """Validates the value is a function, and if so, returns it."""
        if not callable(value):
            raise QuestionCreationValidationError(
                f"Expected a function (got {value}).)"
            )
        return value


class IntegerDescriptor(BaseDescriptor):
    """
    Validates that a value is an integer.
    - `none_allowed` is whether None is allowed as a value.
    """

    def __init__(self, none_allowed: bool = False):
        self.none_allowed = none_allowed

    def validate(self, value, instance):
        if self.none_allowed:
            if not (isinstance(value, int) or value is None):
                raise QuestionAnswerValidationError(
                    f"Expected an integer or None (got {value})."
                )
        else:
            if not isinstance(value, int):
                raise QuestionAnswerValidationError(
                    f"Expected an integer (got {value})."
                )


class IntegerOrNoneDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        if not (isinstance(value, int) or value is None):
            raise QuestionCreationValidationError(
                f"Expected an integer or None (got {value})."
            )


class NumericalOrNoneDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        if not is_number_or_none(value):
            raise QuestionAnswerValidationError(
                f"Expected a number or None (got {value})."
            )


################################
# Attribute-specific descriptors
################################


class AllowNonresponseDescriptor(BaseDescriptor):
    """Validates that the `allow_nonresponse` attribute is a boolean."""

    def validate(self, value, instance):
        if not isinstance(value, bool):
            raise QuestionCreationValidationError(
                f"`allow_nonresponse` must be a boolean (got {value})."
            )


class AnswerTemplateDescriptor(BaseDescriptor):
    """Validates that the answer template is a dictionary with string keys and string values."""

    def validate(self, value: Any, instance) -> None:
        if not isinstance(value, dict):
            raise QuestionCreationValidationError(
                f"`answer_template` must be a dictionary (got {value}).)"
            )
        if not all(isinstance(x, str) for x in value.keys()):
            raise QuestionCreationValidationError(
                f"`answer_template` keys must be strings (got {value})."
            )


class InstructionsDescriptor(BaseDescriptor):
    """Validates that the `instructions` attribute is a string."""

    def validate(self, value, instance):
        # if not isinstance(value, str):
        #     raise QuestionCreationValidationError(
        #         f"Question `instructions` must be a string (got {value})."
        #     )
        pass


class NumSelectionsDescriptor(BaseDescriptor):
    """Validates that `num_selections` is an integer, is less than the number of options, and is positive."""

    def validate(self, value, instance):
        if not (isinstance(value, int)):
            raise QuestionCreationValidationError(
                f"`num_selections` must be an integer (got {value})."
            )
        if value > len(instance.question_options):
            raise QuestionAnswerValidationError(
                f"`num_selections` must be less than the number of options (got {value})."
            )
        if value < 1:
            raise QuestionAnswerValidationError(
                f"`num_selections` must a positive integer (got {value})."
            )


class OptionLabelDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        if value is not None:
            if min(value.keys()) != min(instance.question_options):
                raise QuestionCreationValidationError(
                    f"First option needs a label (got {value})"
                )
            if max(value.keys()) != max(instance.question_options):
                raise QuestionCreationValidationError(
                    f"Last option needs a label (got {value})"
                )
            if not all(isinstance(x, str) for x in value.values()):
                raise QuestionCreationValidationError(
                    "Option labels must be strings (got {value})."
                )
            for key in value.keys():
                if key not in instance.question_options:
                    raise QuestionCreationValidationError(
                        f"Option label key ({key}) is not in question options ({instance.question_options})."
                    )


class QuestionNameDescriptor(BaseDescriptor):
    """Validates that the `question_name` attribute is a valid variable name."""

    def validate(self, value, instance):
        if not is_valid_variable_name(value):
            raise QuestionCreationValidationError(
                f"`question_name` is not a valid variable name (got {value})."
            )


class QuestionOptionsDescriptor(BaseDescriptor):
    """Validates that `question_options` is a list, does not exceed the min/max lengths, and has unique items."""

    def __init__(self, num_choices: int = None, linear_scale: bool = False):
        self.num_choices = num_choices
        self.linear_scale = linear_scale

    def validate(self, value: Any, instance) -> None:
        if not isinstance(value, list):
            raise QuestionCreationValidationError(
                f"Question options must be a list (got {value})."
            )
        if len(value) > Settings.MAX_NUM_OPTIONS:
            raise QuestionCreationValidationError(
                f"Too many question options (got {value})."
            )
        if len(value) < Settings.MIN_NUM_OPTIONS:
            raise QuestionCreationValidationError(
                f"Too few question options (got {value})."
            )
        if len(value) != len(set(value)):
            raise QuestionCreationValidationError(
                f"Question options must be unique (got {value})."
            )
        if not self.linear_scale:
            if not all(isinstance(x, str) for x in value):
                raise QuestionCreationValidationError(
                    "Question options must be strings (got {value}).)"
                )
            if not all(
                [
                    len(option) >= 1 and len(option) < Settings.MAX_OPTION_LENGTH
                    for option in value
                ]
            ):
                raise QuestionCreationValidationError(
                    f"All question options must be at least 1 character long but less than {Settings.MAX_OPTION_LENGTH} characters long (got {value})."
                )

        if hasattr(instance, "min_selections") and instance.min_selections != None:
            if instance.min_selections > len(value):
                raise QuestionCreationValidationError(
                    f"You asked for at least {instance.min_selections} selections, but provided fewer options (got {value})."
                )
        if hasattr(instance, "max_selections") and instance.max_selections != None:
            if instance.max_selections > len(value):
                raise QuestionCreationValidationError(
                    f"You asked for at most {instance.max_selections} selections, but provided fewer options (got {value})."
                )
        if self.num_choices is not None:
            if len(value) != self.num_choices:
                raise QuestionCreationValidationError(
                    f"You asked for {self.num_choices} selections, but provided {len(value)} options."
                )
        if self.linear_scale:
            if sorted(value) != list(range(min(value), max(value) + 1)):
                raise QuestionCreationValidationError(
                    f"LinearScale.question_options must be a list of successive integers, e.g. [1, 2, 3] (got {value})."
                )


class QuestionTextDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        if len(value) > Settings.MAX_QUESTION_LENGTH:
            raise Exception("Question is too long!")
        if len(value) < 1:
            raise Exception("Question is too short!")
        if not isinstance(value, str):
            raise Exception("Question must be a string!")
        if contains_single_braced_substring(value):
            print(
                f"WARNING: Question text contains a single-braced substring: {value}.\nYou probably mean to use a double-braced substring, e.g. {{variable}}."
            )


class ShortNamesDictDescriptor(BaseDescriptor):
    def validate(self, value, instance):
        "Validates the short names dictionary"
        if not isinstance(value, dict):
            raise QuestionCreationValidationError(
                f"Short names dictionary must be a dictionary (got {value})."
            )
        if not all(isinstance(x, str) for x in value.keys()):
            raise QuestionCreationValidationError(
                f"Short names dictionary keys must be strings (got {value})."
            )
        if not all(isinstance(x, str) for x in value.values()):
            raise QuestionCreationValidationError(
                f"Short names dictionary values must be strings (got {value})."
            )



from __future__ import annotations
from collections import UserList
from typing import Optional, Union

from rich.table import Table

from edsl.scenarios.Scenario import Scenario
from edsl.Base import Base


class ScenarioList(Base, UserList):
    def __init__(self, data: Optional[list] = None):
        if data is not None:
            super().__init__(data)

    def to(self, question_or_survey: Union["Question", "Survey"]):
        return question_or_survey.by(*self)

    @classmethod
    def from_csv(cls, filename):
        import csv

        observations = []
        with open(filename, "r") as f:
            reader = csv.reader(f)
            header = next(reader)
            for row in reader:
                observations.append(Scenario(dict(zip(header, row))))
        return cls(observations)

    def to_dict(self):
        return {"scenarios": [s.to_dict() for s in self]}

    @classmethod
    def gen(cls, scenario_dicts_list):
        return cls([Scenario(s) for s in scenario_dicts_list])

    @classmethod
    def from_dict(cls, data):
        return cls([Scenario.from_dict(s) for s in data["scenarios"]])

    def code(self):
        ## TODO: Refactor to only use the questions actually in the survey
        "Creates the Python code representation of a survey"
        header_lines = [
            "from edsl.scenarios.Scenario import Scenario",
            "from edsl.scenarios.ScenarioList import ScenarioList",
        ]
        lines = ["\n".join(header_lines)]
        names = []
        for index, scenario in enumerate(self):
            lines.append(f"scenario_{index} = " + repr(scenario))
            names.append(f"scenario_{index}")
        lines.append(f"scenarios = ScenarioList([{', '.join(names)}])")
        return lines

    @classmethod
    def example(cls):
        return cls([Scenario.example(), Scenario.example()])

    def rich_print(self):
        """Displays an object as a table."""
        table = Table(title="ScenarioList")
        table.add_column("Index", style="bold")
        table.add_column("Scenario")
        for i, s in enumerate(self):
            table.add_row(str(i), s.rich_print())
        return table


if __name__ == "__main__":
    from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
    from edsl.scenarios.Scenario import Scenario

    q = QuestionMultipleChoice(
        question_text="Do you enjoy the taste of {{food}}?",
        question_options=["Yes", "No"],
        question_name="food_preference",
    )

    scenario_list = ScenarioList(
        [Scenario({"food": "wood chips"}), Scenario({"food": "wood-fired pizza"})]
    )

    print(scenario_list.code())

    results = scenario_list.to(q).run()
    print(results)



from __future__ import annotations
import asyncio
import logging
import textwrap
from collections import UserDict
from collections import defaultdict

from typing import Any, Type, List, Generator, Callable
from edsl.agents import Agent
from edsl.language_models import LanguageModel
from edsl.questions import Question
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.utilities.decorators import sync_wrapper
from edsl.data_transfer_models import AgentResponseDict
from edsl.jobs.Answers import Answers
from collections import UserList

from typing import Dict, List

from edsl.exceptions.agents import FailedTaskException


class InterviewError(Exception):
    """Base class for exceptions in this module."""

    pass


class InterviewErrorPriorTaskCanceled(InterviewError):
    """Raised when a prior task was canceled."""

    pass


class InterviewTimeoutError(InterviewError):
    pass


# create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.propagate = False
# create  file handler
fh = logging.FileHandler("interview.log")
fh.setLevel(logging.INFO)
# add formatter to the handlers
formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(module)s:%(lineno)d - %(funcName)s - %(message)s"
)
fh.setFormatter(formatter)
# add handler to logger
logger.addHandler(fh)

# start loggin'
logger.info("Interview.py loaded")

from edsl.config import Config

TIMEOUT = int(Config().API_CALL_TIMEOUT_SEC)


class FailedTask(UserDict):
    def __init__(self, e: Exception = None):
        data = {
            "answer": "Failure",
            "comment": "Failure",
            "prompts": {"user_prompt": "", "sytem_prompt": ""},
            "exception": e,
        }
        super().__init__(data)


class QuestionTaskCreator(UserList):
    """Class to create and manage question tasks with dependencies."""

    def __init__(self, func: Callable):
        # self.func = func  # the function to be called to actually run the task
        super().__init__([])
        self.func = func

    async def _run_task_async(self, question, debug) -> asyncio.Task:
        """Runs the task asynchronously, awaiting the tasks that must be completed before this one can be run."""
        logger.info(f"Running task for {question.question_name}")
        try:
            # This is waiting for the tasks that must be completed before this one can be run.
            # This does *not* use the return_exceptions = True flag, so if any of the tasks fail,
            # it throws the exception immediately, which is what we want.
            await asyncio.gather(*self)
        except Exception as e:
            logger.error(f"Required tasks for {question.question_name} failed: {e}")
            # turns the parent exception into a custom exception
            # So the task gets canceled but this InterviewErrorPriorTaskCanceled exception
            # So we never get the question details we need.
            raise InterviewErrorPriorTaskCanceled(
                f"Required tasks failed for {question.question_name}"
            ) from e
        else:
            logger.info(f"Tasks for {question.question_name} completed")
            # This is the actual task that we want to run.
            results = await self.func(question, debug)
            return results

    def __call__(self, question, debug):
        """Creates a task that depends on the passed-in dependencies."""
        task = asyncio.create_task(self._run_task_async(question, debug))
        task.edsl_name = question.question_name
        return task


class Interview:
    """
    A class that has an Agent answer Survey Questions with a particular Scenario and using a LanguageModel.
    """

    def __init__(
        self,
        agent: Agent,
        survey: Survey,
        scenario: Scenario,
        model: Type[LanguageModel],
        verbose: bool = False,
        debug: bool = False,
    ):
        self.agent = agent
        self.survey = survey
        self.scenario = scenario
        self.model = model
        self.debug = debug
        self.verbose = verbose
        self.answers: dict[str, str] = Answers()  # will get filled in

        self.dag = self.survey.dag(
            textify=True
        )  # the DAG tells us what questions depend on what other questions
        self.to_index = {
            name: index for index, name in enumerate(self.survey.question_names)
        }

        logger.info(f"Interview instantiated")

    async def async_conduct_interview(
        self, debug: bool = False, replace_missing: bool = True
    ) -> tuple["Answers", List[dict[str, Any]]]:
        """
        Conducts an 'interview' asynchronously. An interview is:
        - one agent
        - one survey (so multiple questions)
        - one model
        - one scenario

        Args:
            debug (bool): Enable debugging mode.
            replace_missing (bool): Replace missing answers with None.
            threaded (bool): Flag to use threading if required.

        Returns:
            Tuple[Answers, List[Dict[str, Any]]]: The answers and a list of valid results.
        """

        # Each question ===> 1 task; this creates a list of tasks
        # the same length as the number of questions in the survey
        tasks, invigilators = self._build_question_tasks(debug)
        # when return_exceptions=False, it will just raise the exception
        # and break the loop; otherwise it returns.
        await asyncio.gather(*tasks, return_exceptions=True)

        if replace_missing:
            self.answers.replace_missing_answers_with_none(self.survey)

        valid_results = list(self._extract_valid_results(tasks, invigilators))

        logger.info(f"Total of tasks requested:\t {len(tasks)}")
        logger.info(f"Number of valid results:\t {len(valid_results)}")
        return self.answers, valid_results

    conduct_interview = sync_wrapper(async_conduct_interview)

    def _extract_valid_results(
        self, tasks, invigialtors
    ) -> Generator["Answers", None, None]:
        """Extracts the valid results from the list of results."""

        warning_header = textwrap.dedent(
            """\
            WARNING: At least one question in the survey was not answered.
            """
        )
        assert len(tasks) == len(invigialtors)
        warning_printed = False

        for task, invigilator in zip(tasks, invigialtors):
            if task.done():
                if (exception := task.exception()) is not None:
                    if not warning_printed:
                        warning_printed = True
                        print(warning_header)
                    print(
                        f"""Task `{task.edsl_name}` failed with `{exception.__class__.__name__}`:`{exception}`."""
                    )
                    # if task failed, we use the invigilator to get the failed task result
                    result = invigilator.get_failed_task_result()
                else:
                    result = task.result()
                yield result

    def _build_question_tasks(self, debug) -> List[asyncio.Task]:
        """Creates a task for each question, with dependencies on the questions that must be answered before this one can be answered."""
        logger.info("Creating tasks for each question")
        tasks = []
        invigilators = []
        for question in self.survey.questions:
            tasks_that_must_be_completed_before = (
                self._get_tasks_that_must_be_completed_before(tasks, question)
            )
            question_task = self._create_question_task(
                question, tasks_that_must_be_completed_before, debug
            )
            tasks.append(question_task)
            invigilators.append(self.get_invigilator(question, debug))
        return tasks, invigilators

    def _get_tasks_that_must_be_completed_before(
        self, tasks, question
    ) -> List[asyncio.Task]:
        """Returns the tasks that must be completed before the given question can be answered.
        If a question has no dependencies, this will be an empty list, [].
        """
        parents_of_focal_question = self.dag.get(question.question_name, [])
        return [
            tasks[self.to_index[parent_question_name]]
            for parent_question_name in parents_of_focal_question
        ]

    def _create_question_task(
        self,
        question: Question,
        tasks_that_must_be_completed_before: List[asyncio.Task],
        debug,
    ):
        """Creates a task that depends on the passed-in dependencies that are awaited before the task is run.
        The key awaitable is the `run_task` function, which is a wrapper around the `answer_question_and_record_task` method.
        """
        # creates the *function* that creates the task
        create_task = QuestionTaskCreator(func=self._answer_question_and_record_task)
        for dependency in tasks_that_must_be_completed_before:
            create_task.append(dependency)
        return create_task(question, debug)

    def async_timeout_handler(timeout):
        def decorator(func):
            async def wrapper(*args, **kwargs):
                try:
                    return await asyncio.wait_for(func(*args, **kwargs), timeout)
                except asyncio.TimeoutError:
                    raise InterviewTimeoutError(
                        f"Task timed out after {timeout} seconds."
                    )

            return wrapper

        return decorator

    def get_invigilator(self, question, debug):
        invigilator = self.agent.create_invigilator(
            question=question,
            scenario=self.scenario,
            model=self.model,
            debug=debug,
            memory_plan=self.survey.memory_plan,
            current_answers=self.answers,
        )
        return invigilator

    @async_timeout_handler(TIMEOUT)
    async def _answer_question_and_record_task(
        self,
        question,
        debug,
    ) -> AgentResponseDict:
        """Answers a question and records the task.
        This in turn calls the the passed-in agent's async_answer_question method, which returns a response dictionary.
        """
        invigilator = self.get_invigilator(question, debug=debug)
        response: AgentResponseDict = await invigilator.async_answer_question()
        response["question_name"] = question.question_name
        self.answers.add_answer(response, question)
        return response

    #######################
    # Dunder methods
    #######################
    def __repr__(self) -> str:
        """Returns a string representation of the Interview instance."""
        return f"Interview(agent = {self.agent}, survey = {self.survey}, scenario = {self.scenario}, model = {self.model})"


def main():
    from edsl.language_models import LanguageModelOpenAIThreeFiveTurbo
    from edsl.agents import Agent
    from edsl.surveys import Survey
    from edsl.scenarios import Scenario
    from edsl.questions import QuestionMultipleChoice

    # from edsl.jobs.Interview import Interview

    #  a survey with skip logic
    q0 = QuestionMultipleChoice(
        question_text="Do you like school?",
        question_options=["yes", "no"],
        question_name="q0",
    )
    q1 = QuestionMultipleChoice(
        question_text="Why not?",
        question_options=["killer bees in cafeteria", "other"],
        question_name="q1",
    )
    q2 = QuestionMultipleChoice(
        question_text="Why?",
        question_options=["**lack*** of killer bees in cafeteria", "other"],
        question_name="q2",
    )
    s = Survey(questions=[q0, q1, q2])
    s = s.add_rule(q0, "q0 == 'yes'", q2)

    # create an interview
    a = Agent(traits=None)
    scenario = Scenario()
    m = LanguageModelOpenAIThreeFiveTurbo(use_cache=True)
    I = Interview(agent=a, survey=s, scenario=scenario, model=m)

    # # conduct five interviews
    # for _ in range(5):
    #     I.conduct_interview(debug=True)

    # # replace missing answers
    # I
    # repr(I)
    # eval(repr(I))


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
import asyncio
import json
from typing import Coroutine, Dict, Any
from collections import UserDict

from edsl.exceptions import AgentRespondedWithBadJSONError
from edsl.prompts.Prompt import Prompt
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler
from edsl.prompts.registry import get_classes
from edsl.exceptions import QuestionScenarioRenderError

from edsl.data_transfer_models import AgentResponseDict


from edsl.exceptions.agents import FailedTaskException


class InvigilatorBase(ABC):
    """An invigiator is a class that is responsible for administering a question to an agent."""

    def __init__(self, agent, question, scenario, model, memory_plan, current_answers):
        self.agent = agent
        self.question = question
        self.scenario = scenario
        self.model = model
        self.memory_plan = memory_plan
        self.current_answers = current_answers

    def get_failed_task_result(self):
        return AgentResponseDict(
            answer=None,
            comment="Failed to get response",
            question_name=self.question.question_name,
            prompts=self.get_prompts(),
        )

    def get_prompts(self) -> Dict[str, Prompt]:
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }

    @classmethod
    def example(cls):
        """Returns an example invigilator."""
        from edsl.agents.Agent import Agent
        from edsl.questions import QuestionMultipleChoice
        from edsl.scenarios.Scenario import Scenario
        from edsl.language_models import LanguageModel

        from edsl.enums import LanguageModelType, InferenceServiceType

        class TestLanguageModelGood(LanguageModel):
            _model_ = LanguageModelType.TEST.value
            _parameters_ = {"temperature": 0.5}
            _inference_service_ = InferenceServiceType.TEST.value

            async def async_execute_model_call(
                self, user_prompt: str, system_prompt: str
            ) -> dict[str, Any]:
                await asyncio.sleep(0.1)
                return {"message": """{"answer": "SPAM!"}"""}

            def parse_response(self, raw_response: dict[str, Any]) -> str:
                return raw_response["message"]

        model = TestLanguageModelGood()
        agent = Agent.example()
        question = QuestionMultipleChoice.example()
        scenario = Scenario.example()
        #        model = LanguageModel.example()
        memory_plan = None
        current_answers = None
        return cls(
            agent=agent,
            question=question,
            scenario=scenario,
            model=model,
            memory_plan=memory_plan,
            current_answers=current_answers,
        )

    @abstractmethod
    async def async_answer_question(self):
        "This is the async method that actually answers the question."
        pass

    @jupyter_nb_handler
    def answer_question(self) -> Coroutine:
        async def main():
            results = await asyncio.gather(self.async_answer_question())
            return results[0]  # Since there's only one task, return its result

        return main()

    def create_memory_prompt(self, question_name):
        """Creates a memory for the agent."""
        return self.memory_plan.get_memory_prompt_fragment(
            question_name, self.current_answers
        )


class InvigilatorAI(InvigilatorBase):
    async def async_answer_question(self, failed=False) -> AgentResponseDict:
        data = {
            "agent": self.agent,
            "question": self.question,
            "scenario": self.scenario,
        }
        raw_response = await self.async_get_response(**self.get_prompts())
        response = self._format_raw_response(**(data | {"raw_response": raw_response}))
        return response

    async def async_get_response(self, user_prompt: Prompt, system_prompt: Prompt):
        """Calls the LLM and gets a response. Used in the `answer_question` method."""
        try:
            response = await self.model.async_get_response(
                user_prompt.text, system_prompt.text
            )
        except json.JSONDecodeError as e:
            raise AgentRespondedWithBadJSONError(
                f"Returned bad JSON: {e}"
                f"Prompt: {user_prompt}"
                f"System Prompt: {system_prompt}"
            )

        return response

    def _format_raw_response(
        self, agent, question, scenario, raw_response
    ) -> AgentResponseDict:
        response = question.validate_answer(raw_response)
        comment = response.get("comment", "")
        answer_code = response["answer"]
        answer = question.translate_answer_code_to_answer(answer_code, scenario)
        data = {
            "answer": answer,
            "comment": comment,
            "question_name": question.question_name,
            "prompts": {k: v.to_dict() for k, v in self.get_prompts().items()},
        }
        return data

    get_response = sync_wrapper(async_get_response)

    def construct_system_prompt(self) -> Prompt:
        """Constructs the system prompt for the LLM call."""
        applicable_prompts = get_classes(
            component_type="agent_instructions",
            model=self.model.model,
        )
        if len(applicable_prompts) == 0:
            raise Exception("No applicable prompts found")

        agent_instructions = applicable_prompts[0](text=self.agent.instruction)

        if not hasattr(self.agent, "agent_persona"):
            applicable_prompts = get_classes(
                component_type="agent_persona",
                model=self.model.model,
            )
            persona_prompt_template = applicable_prompts[0]()
        else:
            persona_prompt_template = self.agent.agent_persona

        if undefined := persona_prompt_template.undefined_template_variables(
            self.agent.traits
            | {"traits": self.agent.traits}
            | {"codebook": self.agent.codebook}
            | {"traits": self.agent.traits}
        ):
            raise QuestionScenarioRenderError(
                f"Agent persona still has variables that were not rendered: {undefined}"
            )

        persona_prompt = persona_prompt_template.render(
            self.agent.traits | {"traits": self.agent.traits},
            codebook=self.agent.codebook,
            traits=self.agent.traits,
        )

        if persona_prompt.has_variables:
            raise QuestionScenarioRenderError(
                "Agent persona still has variables that were not rendered."
            )

        return (
            agent_instructions
            + " " * int(len(persona_prompt.text) > 0)
            + persona_prompt
        )

    def get_question_instructions(self) -> Prompt:
        """Gets the instructions for the question."""
        applicable_prompts = get_classes(
            component_type="question_instructions",
            question_type=self.question.question_type,
            model=self.model.model,
        )
        ## Get the question instructions and renders with the scenario & question.data
        question_prompt = applicable_prompts[0]()

        undefined_template_variables = question_prompt.undefined_template_variables(
            self.question.data | self.scenario
        )
        if undefined_template_variables:
            print(undefined_template_variables)
            raise QuestionScenarioRenderError(
                "Question instructions still has variables"
            )

        return question_prompt.render(self.question.data | self.scenario)

    def construct_user_prompt(self) -> Prompt:
        """Gets the user prompt for the LLM call."""
        user_prompt = self.get_question_instructions()
        if self.memory_plan is not None:
            user_prompt += self.create_memory_prompt(self.question.question_name)
        return user_prompt

    def get_prompts(self) -> Dict[str, Prompt]:
        """Gets the prompts for the LLM call."""
        system_prompt = self.construct_system_prompt()
        user_prompt = self.construct_user_prompt()
        return {
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }

    answer_question = sync_wrapper(async_answer_question)


class InvigilatorDebug(InvigilatorBase):
    async def async_answer_question(self) -> AgentResponseDict:
        results = self.question.simulate_answer(human_readable=True)
        results["prompts"] = self.get_prompts()
        results["question_name"] = self.question.question_name
        return AgentResponseDict(**results)

    def get_prompts(self) -> Dict[str, Prompt]:
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }


class InvigilatorHuman(InvigilatorBase):
    async def async_answer_question(self) -> AgentResponseDict:
        data = {
            "comment": "This is a real survey response from a human.",
            "answer": None,
            "prompts": self.get_prompts(),
            "question_name": self.question.question_name,
        }
        try:
            answer = self.agent.answer_question_directly(self.question, self.scenario)
            return AgentResponseDict(**(data | {"answer": answer}))
        except Exception as e:
            agent_response_dict = AgentResponseDict(
                **(data | {"answer": None, "comment": str(e)})
            )
            raise FailedTaskException(
                "Failed to get response", agent_response_dict
            ) from e


class InvigilatorFunctional(InvigilatorBase):
    async def async_answer_question(self) -> AgentResponseDict:
        func = self.question.answer_question_directly
        data = {
            "comment": "Functional.",
            "prompts": self.get_prompts(),
            "question_name": self.question.question_name,
        }
        try:
            answer = func(scenario=self.scenario, agent_traits=self.agent.traits)
            return AgentResponseDict(**(data | {"answer": answer}))
        except Exception as e:
            agent_response_dict = AgentResponseDict(
                **(data | {"answer": None, "comment": str(e)})
            )
            raise FailedTaskException(
                "Failed to get response", agent_response_dict
            ) from e

    def get_prompts(self) -> Dict[str, Prompt]:
        return {
            "user_prompt": Prompt("NA").text,
            "system_prompt": Prompt("NA").text,
        }


if __name__ == "__main__":
    from edsl.enums import LanguageModelType

    from edsl.agents.Agent import Agent

    a = Agent(
        instruction="You are a happy-go lucky agent.",
        traits={"feeling": "happy", "age": "Young at heart"},
        codebook={"feeling": "Feelings right now", "age": "Age in years"},
        trait_presentation_template="",
    )

    class MockModel:
        model = LanguageModelType.GPT_4.value

    class MockQuestion:
        question_type = "free_text"
        question_text = "How are you feeling?"
        question_name = "feelings_question"
        data = {
            "question_name": "feelings",
            "question_text": "How are you feeling?",
            "question_type": "feelings_question",
        }

    i = InvigilatorAI(
        agent=a,
        question=MockQuestion(),
        scenario={},
        model=MockModel(),
        memory_plan=None,
        current_answers=None,
    )
    print(i.get_prompts()["system_prompt"])
    assert i.get_prompts()["system_prompt"].text == "You are a happy-go lucky agent."

    ###############
    ## Render one
    ###############

    a = Agent(
        instruction="You are a happy-go lucky agent.",
        traits={"feeling": "happy", "age": "Young at heart"},
        codebook={"feeling": "Feelings right now", "age": "Age in years"},
        trait_presentation_template="You are feeling {{ feeling }}.",
    )

    i = InvigilatorAI(
        agent=a,
        question=MockQuestion(),
        scenario={},
        model=MockModel(),
        memory_plan=None,
        current_answers=None,
    )
    print(i.get_prompts()["system_prompt"])

    assert (
        i.get_prompts()["system_prompt"].text
        == "You are a happy-go lucky agent. You are feeling happy."
    )
    try:
        assert i.get_prompts()["system_prompt"].unused_traits(a.traits) == ["age"]
    except AssertionError:
        unused_traits = i.get_prompts()["system_prompt"].unused_traits(a.traits)
        print(f"System prompt: {i.get_prompts()['system_prompt']}")
        print(f"Agent traits: {a.traits}")
        print(f"Unused_traits: {unused_traits}")
        # breakpoint()

    ###############
    ## Render one
    ###############

    a = Agent(
        instruction="You are a happy-go lucky agent.",
        traits={"feeling": "happy", "age": "Young at heart"},
        codebook={"feeling": "Feelings right now", "age": "Age in years"},
        trait_presentation_template="You are feeling {{ feeling }}. You eat lots of {{ food }}.",
    )

    i = InvigilatorAI(
        agent=a,
        question=MockQuestion(),
        scenario={},
        model=MockModel(),
        memory_plan=None,
        current_answers=None,
    )
    print(i.get_prompts()["system_prompt"])

    ## Should raise a QuestionScenarioRenderError
    assert (
        i.get_prompts()["system_prompt"].text
        == "You are a happy-go lucky agent. You are feeling happy."
    )



from dataclasses import dataclass, field


@dataclass
class BaseDataClass:
    example_data: dict = field(default_factory=dict)
    short_names_dict: dict = field(default_factory=dict)

    @classmethod
    def example(cls):
        data = cls.example_data
        return cls(**data)


@dataclass
class ChiSquareData:
    chi_square: float
    p_value: float
    text: str
    example_data: dict = field(
        default_factory=lambda: {
            "chi_square": 1.0,
            "p_value": 0.5,
            "text": "Debug",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class RegressionData:
    model_outcome: str
    outcome_description: str
    example_data: dict = field(
        default_factory=lambda: {
            "model_outcome": "Debug",
            "outcome_description": "Debug",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class FreeTextData:
    responses: list[str]
    text: str
    example_data: dict = field(
        default_factory=lambda: {
            "responses": ["Bad dog", "Bad cat", "Great chicken", "Great cow"],
            "text": "Debug",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class CategoricalData:
    responses: list[str]
    options: list[str]
    text: str
    example_data: dict = field(
        default_factory=lambda: {
            "responses": ["Bad", "Bad", "Great", "Great"],
            "options": ["Good", "Great", "OK", "Bad"],
            "text": "Debug",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class NumericalData:
    responses: list[float]
    text: str
    example_data: dict = field(
        default_factory=lambda: {"responses": [1, 2, 3, 4], "text": "Debug"}
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class TallyData:
    responses: dict
    text: str
    example_data: dict = field(
        default_factory=lambda: {
            "responses": {"Bad": 2, "Great": 2, "Good": 0, "OK": 0},
            "text": "Debug",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class CrossTabData:
    cross_tab: dict
    left_title: str
    right_title: str
    example_data: dict = field(
        default_factory=lambda: {
            "cross_tab": {
                "Bad": {"Bad": 2, "Great": 0, "Good": 0, "OK": 0},
                "Great": {"Bad": 0, "Great": 2, "Good": 0, "OK": 0},
                "Good": {"Bad": 0, "Great": 0, "Good": 0, "OK": 0},
                "OK": {"Bad": 0, "Great": 0, "Good": 0, "OK": 0},
            },
            "left_title": "Debug left question",
            "right_title": "Debug right question",
        }
    )
    short_names_dict: dict = field(default_factory=dict)


@dataclass
class PlotData:
    buffer: bytes
    title: str
    option_codes: dict[str, str]
    width_pct: int
    example_data: dict = field(
        default_factory=lambda: {
            "buffer": b"123",
            "title": "Debug",
            "option_codes": {},
            "width_pct": 100,
        }
    )
    short_names_dict: dict = field(default_factory=dict)


if __name__ == "__main__":
    print(CategoricalData.example())
    print(NumericalData.example())



from collections import UserDict
import importlib

from edsl.jobs.runners.JobsRunnerAsyncio import JobsRunnerAsyncio
from edsl.jobs.runners.JobsRunnerDryRun import JobsRunnerDryRun

from edsl.exceptions import JobsRunError
from edsl.jobs.JobsRunner import RegisterJobsRunnerMeta


class JobsRunnersRegistryDict(UserDict):
    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise JobsRunError(f"JobsRunner '{key}' not found in registry.")


registry_data = RegisterJobsRunnerMeta.lookup()
JobsRunnersRegistry = JobsRunnersRegistryDict(registry_data)


class JobsRunnerDescriptor:
    def validate(self, value: str) -> None:
        """Validates the value. If it is invalid, raises an exception. If it is valid, does nothing."""
        if value not in JobsRunnersRegistry:
            raise ValueError(
                f"JobsRunner must be one of {list(JobsRunnersRegistry.keys())}"
            )

    def __get__(self, instance, owner):
        """"""
        if self.name not in instance.__dict__:
            return None
        else:
            return instance.__dict__[self.name]

    def __set__(self, instance, value: str) -> None:
        self.validate(value, instance)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name: str) -> None:
        self.name = "_" + name


if __name__ == "__main__":
    pass



from sqlalchemy import desc
from typing import Union
from edsl.data import Database, database, LLMOutputDataDB
from edsl.data.orm import ResultDB


class CRUDOperations:
    """
    A class that implementes CRUD operations for the EDSL package.

    Initalization:
    - `database`: A Database object.

    Methods:
    - `get_LLMOutputData(model, parameters, system_prompt, prompt)`: Retrieves a cached LLM output from the database.
    - `write_LLMOutputData(model, parameters, system_prompt, prompt, output)`: Writes an LLM output to the database.
    """

    def __init__(self, database: Database) -> None:
        self.database = database

    def get_LLMOutputData(
        self, model: str, parameters: str, system_prompt: str, prompt: str
    ) -> Union[str, None]:
        """
        Retrieves a cached LLM output from the database. Arguments: in string format, the model, parameters, system_prompt, and prompt used to generate the output. Returns the output (json string) if it exists, otherwise None.
        """
        with self.database.get_db() as db:
            record = (
                db.query(LLMOutputDataDB)
                .filter_by(
                    prompt=prompt,
                    system_prompt=system_prompt,
                    model=model,
                    parameters=parameters,
                )
                .order_by(desc(LLMOutputDataDB.id))
                .first()
            )
        return record.output if record else None

    def write_LLMOutputData(
        self, model: str, parameters: str, system_prompt: str, prompt: str, output: str
    ) -> None:
        """
        Writes an LLM output to the database. Arguments: in string format, the model, parameters, system_prompt, prompt, and the generated output.
        """
        record = LLMOutputDataDB(
            model=model,
            parameters=parameters,
            system_prompt=system_prompt,
            prompt=prompt,
            output=output,
        )

        with self.database.get_db() as db:
            db.add(record)
            db.commit()

    def write_result(
        self,
        job_uuid: str,
        result_uuid: str,
        agent: str,
        scenario: str,
        model: str,
        answer: str,
    ) -> None:
        """Writes a Result record to the database."""
        record = ResultDB(
            job_uuid=job_uuid,
            result_uuid=result_uuid,
            agent=agent,
            scenario=scenario,
            model=model,
            answer=answer,
        )

        with self.database.get_db() as db:
            db.add(record)
            db.commit()

    def read_results(self, job_uuid: str) -> list[ResultDB]:
        """Reads all Result records associated with job_uuid from the database."""
        with self.database.get_db() as db:
            records = (
                db.query(ResultDB)
                .filter_by(job_uuid=job_uuid)
                .order_by(desc(ResultDB.id))
                .all()
            )
        return records


CRUD = CRUDOperations(database)



class SurveyErrors(Exception):
    pass


class SurveyCreationError(SurveyErrors):
    pass


class SurveyHasNoRulesError(SurveyErrors):
    pass


class SurveyRuleSendsYouBackwardsError(SurveyErrors):
    pass


class SurveyRuleSkipLogicSyntaxError(SurveyErrors):
    pass


class SurveyRuleReferenceInRuleToUnknownQuestionError(SurveyErrors):
    pass


class SurveyRuleRefersToFutureStateError(SurveyErrors):
    pass


class SurveyRuleCollectionHasNoRulesAtNodeError(SurveyErrors):
    pass


class SurveyRuleCannotEvaluateError(SurveyErrors):
    pass



"""
This an ABC for a tracker that monitors a Python queue. 

Contained within the namespace is a class called TrackerEvent, which is also an ABC. 
This is the kind of event that will be placed in the queue the tracker is monitoring.
Each child class of TrackerEvent needs to know how to update the state of the tracker.

Each child has to implement a tracked_values method which tells what values the 
tracker is tracking - this is so it can pretty-print them. 

All child classes have to implement an allowed_events class variable. 

"""
import time
from abc import ABC, abstractmethod
from queue import Empty
from edsl.utilities.interface import print_dict_with_rich


class Tracker(ABC):
    "Meant to be single-threaded"

    class TrackerEvent(ABC):
        "Each child event needs to know how to update the state of the tracker."

        @abstractmethod
        def apply(self, tracker):
            pass

    def __init__(self, event_queue, verbose=False):
        self.event_queue = event_queue
        self.observed_events = []
        self.verbose = verbose
        if self.verbose:
            print("Tracker instantiated")

    @abstractmethod
    def allowed_events(self) -> set:
        "What events (classes) are allowed to be placed in the queue?"
        pass

    @abstractmethod
    def tracked_values(self) -> dict:
        "What values are tracked by this tracker (should be flat dictionary)?"
        pass

    def get_from_queue(self):
        """Gets events from the queue and stores them in a list.
        It runs until the event_queue is empty.
        """
        while True:
            try:
                event = self.event_queue.get(block=False)
            except Empty as e:
                if self.verbose:
                    print("The queue is empty.")
                break
            else:
                self.observed_events.append(event)
                self.process_event(event)
                self.event_queue.task_done()

    def __call__(self, all_done, interval=1):
        if self.verbose:
            print("Tracking thread is starting.")
        while not all_done.is_set():
            self.get_from_queue()
            time.sleep(interval)

        # Run one last time to get anything still left in the event_queue
        self.get_from_queue()

    def process_event(self, event):
        """Processes an event from the event queue."""

        ## diabling this for now as two error-prone:
        # https://chat.openai.com/share/6e0e2ae2-30d7-4281-982a-a69a218c4e76

        # if not isinstance(event, tuple(self.allowed_events())):
        #    print(f"Event type: {type(event).__name__}")
        #    print(f"Allowed events: {self.allowed_events()}")
        #    raise ValueError(f"Invalid event type: {type(event).__name__}")

        event.apply(self)
        if self.verbose:
            self.show_status()

    def show_status(self):
        """Prints the status of the interview manager."""
        data = self.tracked_values()
        print_dict_with_rich(data)



import json
import requests
from edsl import CONFIG
from edsl.jobs import Jobs

api_url = {
    "development": "http://127.0.0.1:8000",
    "production": "https://api.goemeritus.com",
}


class Client:
    def __init__(self, api_key: str = None, run_mode: str = None) -> None:
        self.api_key = api_key or CONFIG.EMERITUS_API_KEY
        self.run_mode = run_mode or CONFIG.EDSL_RUN_MODE

    def __repr__(self):
        return f"Client(api_key='{self.api_key}', run_mode='{self.run_mode}')"

    @property
    def headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_key}"}

    @property
    def url(self) -> str:
        return api_url[self.run_mode]

    # JOBS METHODS edit
    @property
    def jobs(self) -> list:
        """Returns a list of all jobs on the server."""
        url = f"{self.url}/api/v0/jobs"
        response = requests.get(url, headers=self.headers)
        for k in response.json():
            print(k)

    def send_job(self, job: Jobs) -> dict:
        """Sends a job to the server."""
        url = f"{self.url}/api/v0/jobs"
        payload = {"json_string": json.dumps(job.to_dict())}
        response = requests.post(url, json=payload, headers=self.headers)
        return response.json()

    def run_job(self, job_id: str) -> dict:
        """Runs a job on the server."""
        url = f"{self.url}/api/v0/jobs/{job_id}/run"
        response = requests.post(url, headers=self.headers)
        return response.json()

    def get_results(self, results_id: str) -> dict:
        """Returns the results of a job on the server."""
        url = f"{self.url}/api/v0/results/{results_id}"
        response = requests.get(url, headers=self.headers)
        return response.json()


if __name__ == "__main__":

    def create_example_job() -> Jobs:
        from edsl.agents.Agent import Agent
        from edsl.questions import QuestionMultipleChoice
        from edsl.scenarios.Scenario import Scenario
        from edsl.surveys.Survey import Survey

        q = QuestionMultipleChoice(
            question_text="How are you this {{ period }}?",
            question_options=["Good", "Great", "OK", "Terrible"],
            question_name="how_feeling",
        )
        base_survey = Survey(questions=[q])

        job = base_survey.by(
            Scenario({"period": "morning"}), Scenario({"period": "afternoon"})
        ).by(Agent({"status": "Super duper unhappy"}), Agent({"status": "Joyful"}))

        return job

    # start a client
    client = Client(run_mode="development")

    # basics
    client
    client.headers
    client.url

    # check jobs on server
    client.jobs

    # send a job to the server
    job = create_example_job()
    client.send_job(job)

    client.jobs
    client.run_job(job_id="1")

    client.jobs
    client.get_results(results_id="1")



from __future__ import annotations
import json
import io
from collections import UserList, defaultdict

from rich.console import Console

from simpleeval import EvalWithCompoundTypes
from typing import Any, Type, Union
from edsl.exceptions import (
    ResultsBadMutationstringError,
    ResultsColumnNotFoundError,
    ResultsInvalidNameError,
    ResultsMutateError,
)
from edsl.agents import Agent
from edsl.data import CRUD
from edsl.language_models import LanguageModel
from edsl.results.Dataset import Dataset
from edsl.results.Result import Result
from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.utilities import (
    is_gzipped,
    is_valid_variable_name,
    print_public_methods_with_doc,
    shorten_string,
    is_notebook,
)


class Mixins(ResultsExportMixin):
    pass


# These are only made available if the user has installed
# our package from pip with the [report] option
try:
    from edsl.report.RegressionMixin import RegressionMixin

    Mixins = type("Mixins", (RegressionMixin, Mixins), {})
except (ImportError, ModuleNotFoundError):
    pass

try:
    from edsl.report.ResultsFetchMixin import ResultsFetchMixin

    Mixins = type("Mixins", (ResultsFetchMixin, Mixins), {})
except (ImportError, ModuleNotFoundError):
    pass

try:
    from edsl.report.ResultsOutputMixin import ResultsOutputMixin

    Mixins = type("Mixins", (ResultsOutputMixin, Mixins), {})
except (ImportError, ModuleNotFoundError):
    pass

from edsl.Base import Base


class Results(UserList, Mixins, Base):
    """
    This class is a UserList of Result objects.
    - It is instantiated with a Survey and a list of Result objects (observations).
    - It can be manipulated in various ways with select, filter, mutate, etc.
    - It also has a list of created_columns, which is a list of columns that have been created with `mutate`
    """

    def __init__(
        self,
        survey: Survey = None,
        data: list[Result] = None,
        created_columns: list = None,
        job_uuid: str = None,
        total_results: int = None,
    ):
        super().__init__(data)
        self.survey = survey
        self.created_columns = created_columns or []
        self._job_uuid = job_uuid
        self._total_results = total_results

        if hasattr(self, "add_output_functions"):
            self.add_output_functions()

    ######################
    # Streaming methods
    ######################

    def code(self):
        raise NotImplementedError

    def __getitem__(self, i):
        if isinstance(i, slice):
            # Return a sliced view of the list
            return self.__class__(survey=self.survey, data=self.data[i])
        else:
            # Return a single item
            return self.data[i]

    def _update_results(self) -> None:
        if self._job_uuid and len(self.data) < self._total_results:
            results = [
                Result(
                    agent=Agent.from_dict(json.loads(r.agent)),
                    scenario=Scenario.from_dict(json.loads(r.scenario)),
                    model=LanguageModel.from_dict(json.loads(r.model)),
                    iteration=1,
                    answer=json.loads(r.answer),
                )
                for r in CRUD.read_results(self._job_uuid)
            ]
            self.data = results

    def __repr__(self) -> str:
        self._update_results()
        if self._job_uuid and len(self.data) < self._total_results:
            print(f"Completeness: {len(self.data)}/{self._total_results}")
        data = [repr(result) for result in self.data]
        if is_notebook():
            return self.rich_print()
        else:
            return f"Results(data = {data}, survey = {repr(self.survey)}, created_columns = {self.created_columns})"

    def to_dict(self) -> dict[str, Any]:
        """Converts the Results object to a dictionary"""
        return {
            "data": [result.to_dict() for result in self.data],
            "survey": self.survey.to_dict(),
            "created_columns": self.created_columns,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Results:
        """Converts a dictionary to a Results object"""
        results = cls(
            survey=Survey.from_dict(data["survey"]),
            data=[Result.from_dict(r) for r in data["data"]],
            created_columns=data.get("created_columns", None),
        )
        return results

    def show_methods(self, show_docstrings: bool = True):
        """Prints public methods of the Results class"""
        print_public_methods_with_doc(self)

    ######################
    ## Convenience methods
    ## & Report methods
    ######################
    @property
    def _key_to_data_type(self) -> dict[str, str]:
        """
        Returns a mapping of keys (how_feeling, status, etc.) to strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d = {}
        for result in self.data:
            d.update(result.key_to_data_type)
        for column in self.created_columns:
            d[column] = "answer"
        return d

    @property
    def _data_type_to_keys(self) -> dict[str, str]:
        """
        Returns a mapping of strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.) to keys (how_feeling, status, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d = defaultdict(set)
        for result in self.data:
            for key, value in result.key_to_data_type.items():
                d[value] = d[value].union(set({key}))
        for column in self.created_columns:
            d["answer"] = d["answer"].union(set({column}))
        return d

    @property
    def answer_keys(self) -> dict[str, str]:
        """Returns a mapping of answer keys to question text"""
        answer_keys = self._data_type_to_keys["answer"]
        answer_keys = {k for k in answer_keys if "_comment" not in k}
        questions_text = [
            self.survey.get_question(k).question_text for k in answer_keys
        ]
        short_question_text = [shorten_string(q, 80) for q in questions_text]
        return dict(zip(answer_keys, short_question_text))

    @property
    def agents(self) -> list[Agent]:
        return [r.agent for r in self.data]

    @property
    def models(self) -> list[Type[LanguageModel]]:
        return [r.model for r in self.data]

    @property
    def scenarios(self) -> list[Scenario]:
        return [r.scenario for r in self.data]

    @property
    def agent_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Agent data"""
        return self._data_type_to_keys["agent"]

    @property
    def model_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the LanguageModel data"""
        return self._data_type_to_keys["model"]

    @property
    def scenario_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Scenario data"""
        return self._data_type_to_keys["scenario"]

    @property
    def question_names(self) -> list[str]:
        """Returns a list of all of the question names"""
        if self.survey is None:
            return []
        return list(self.survey.question_names)

    @property
    def all_keys(self) -> set[str]:
        """Returns a set of all of the keys that are in the Results"""
        answer_keys = set(self.answer_keys)
        return (
            answer_keys.union(self.agent_keys)
            .union(self.scenario_keys)
            .union(self.model_keys)
        )

    def relevant_columns(self) -> set[str]:
        """Returns all of the columns that are in the results."""
        return set().union(
            *(observation.combined_dict.keys() for observation in self.data)
        )

    def _parse_column(self, column: str) -> tuple[str, str]:
        """
        Parses a column name into a tuple containing a data type and a key.
        - Uses the key_to_data_type property of the Results class.
        - The standard way a column is specified is with a dot-separated string, e.g. _parse_column("agent.status")
        - But you can also specify a single key, e.g. "status", in which case it will look up the data type.
        """
        if "." in column:
            data_type, key = column.split(".")
        else:
            try:
                data_type, key = self._key_to_data_type[column], column
            except KeyError:
                raise ResultsColumnNotFoundError(f"Column {column} not found in data")
        return data_type, key

    def first(self) -> Result:
        """Returns the first observation in the results."""
        return self.data[0]

    def mutate(self, new_var_string: str, functions_dict: dict = None) -> Results:
        """
        Creates a value in the Results object as if has been asked as part of the survey.
        - It splits the new_var_string at the "=" and uses simple_eval
        - The functions dict is...

        Example:
        >>> r = Results.create_example()
        >>> r.mutate('how_feeling_x = how_feeling + "x"').select('how_feeling_x')
        [{'answer.how_feeling_x': ['Badx', 'Badx', 'Greatx', 'Greatx']}]
        """
        # extract the variable name and the expression
        if "=" not in new_var_string:
            raise ResultsBadMutationstringError(
                f"Mutate requires an '=' in the string, but '{new_var_string}' doesn't have one."
            )
        raw_var_name, expression = new_var_string.split("=", 1)
        var_name = raw_var_name.strip()
        if not is_valid_variable_name(var_name):
            raise ResultsInvalidNameError(f"{var_name} is not a valid variable name.")

        # create the evaluator
        functions_dict = functions_dict or {}

        def create_evaluator(result: Result) -> EvalWithCompoundTypes:
            return EvalWithCompoundTypes(
                names=result.combined_dict, functions=functions_dict
            )

        def new_result(old_result: Result, var_name: str) -> Result:
            evaluator = create_evaluator(old_result)
            value = evaluator.eval(expression)
            new_result = old_result.copy()
            new_result["answer"][var_name] = value
            return new_result

        try:
            new_data = [new_result(result, var_name) for result in self.data]
        except Exception as e:
            raise ResultsMutateError(f"Error in mutate. Exception:{e}")

        return Results(
            survey=self.survey,
            data=new_data,
            created_columns=self.created_columns + [var_name],
        )

    def select(self, *columns: Union[str, list[str]]) -> Dataset:
        """
        This selects data from the results and turns it into a format like so:
        >>> results = Results.create_example()
        >>> results.select('how_feeling')
        [{'answer.how_feeling': ['Bad', 'Bad', 'Great', 'Great']}]
        """

        if not columns or columns == ("*",) or columns == (None,):
            columns = ("*.*",)

        if isinstance(columns[0], list):
            columns = tuple(columns[0])

        known_data_types = ["answer", "scenario", "agent", "model", "prompt"]

        def get_data_types_to_return(parsed_data_type):
            if parsed_data_type == "*":  # they want all of the columns
                return known_data_types
            else:
                if parsed_data_type not in known_data_types:
                    raise Exception(
                        f"Data type {parsed_data_type} not found in data. Did you mean one of {known_data_types}"
                    )
                return [parsed_data_type]

        # we're doing to populate this with the data we want to fetch
        to_fetch = defaultdict(list)

        new_data = []
        items_in_order = []
        # iterate through the passed columns
        for column in columns:
            # a user could pass 'result.how_feeling' or just 'how_feeling'
            parsed_data_type, parsed_key = self._parse_column(column)
            data_types = get_data_types_to_return(parsed_data_type)
            found_once = False  # we need to track this to make sure we found the key at least once

            for data_type in data_types:
                # the keys for that data_type e.g.,# if data_type is 'answer', then the keys are 'how_feeling', 'how_feeling_comment', etc.
                relevant_keys = self._data_type_to_keys[data_type]

                for key in relevant_keys:
                    if key == parsed_key or parsed_key == "*":
                        found_once = True
                        to_fetch[data_type].append(key)
                        items_in_order.append(data_type + "." + key)

            if not found_once:
                raise Exception(f"Key {parsed_key} not found in data.")

        for data_type in to_fetch:
            for key in to_fetch[data_type]:
                entries = self._fetch_list(data_type, key)
                new_data.append({data_type + "." + key: entries})

        def sort_by_key_order(dictionary):
            # Extract the single key from the dictionary
            single_key = next(iter(dictionary))
            # Return the index of this key in the list_of_keys
            return items_in_order.index(single_key)

        sorted(new_data, key=sort_by_key_order)

        return Dataset(new_data)

    def sort_by(self, column, reverse=True) -> Results:
        "Sorts the results by a column"

        data_type, key = self._parse_column(column)

        def to_numeric_if_possible(v):
            try:
                return float(v)
            except:
                return v

        new_data = sorted(
            self.data,
            key=lambda x: to_numeric_if_possible(x.get_value(data_type, key)),
            reverse=reverse,
        )
        return Results(survey=self.survey, data=new_data, created_columns=None)

    def filter(self, expression) -> Results:
        """
        This filters a result based on the expression that is passed in.
        >>> r = Results.create_example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling')
        [{'answer.how_feeling': ['Great', 'Great']}]
        >>> r.filter("how_feeling == 'Nothing'").select('how_feeling')
        [{'answer.how_feeling': []}]
        """

        def create_evaluator(result):
            return EvalWithCompoundTypes(names=result.combined_dict)

        try:
            new_data = [
                result
                for result in self.data
                if create_evaluator(result).eval(expression)
            ]
        except Exception as e:
            print(f"Exception:{e}")

        return Results(survey=self.survey, data=new_data, created_columns=None)

    @classmethod
    def example(cls, debug: bool = False) -> Results:
        """
        Returns an example Results object
        - debug: if False, uses actual API calls
        """
        from edsl.jobs import Jobs

        job = Jobs.example()
        results = job.run()
        return results

    def rich_print(self):
        """Displays an object as a table."""
        with io.StringIO() as buf:
            console = Console(file=buf, record=True)

            for index, result in enumerate(self):
                console.print(f"Result {index}")
                console.print(result.rich_print())

            return console.export_text()

    def __str__(self):
        return self.rich_print()


def main():  # pragma: no cover
    """Calls the OpenAI API credits"""
    from edsl.results.Results import Results

    results = Results.example(debug=True)
    print(results.filter("how_feeling == 'Great'").select("how_feeling"))
    print(results.mutate("how_feeling_x = how_feeling + 'x'").select("how_feeling_x"))


if __name__ == "__main__":
    print(Results.example())



import base64
import csv
import io
from IPython.display import HTML, display
import pandas as pd
from edsl.utilities import (
    print_list_of_dicts_with_rich,
    print_list_of_dicts_as_html_table,
    print_dict_with_rich,
)


class ResultsExportMixin:
    def convert_decorator(func):
        def wrapper(self, *args, **kwargs):
            if self.__class__.__name__ == "Results":
                return func(self.select(), *args, **kwargs)
            elif self.__class__.__name__ == "Dataset":
                return func(self, *args, **kwargs)
            else:
                raise Exception(
                    f"Class {self.__class__.__name__} not recognized as a Results or Dataset object."
                )

        return wrapper

    @convert_decorator
    def _make_tabular(self, remove_prefix):
        "Helper function that turns the results into a tabular format."
        d = {}
        full_header = list(self.relevant_columns())
        for entry in self.data:
            key, list_of_values = list(entry.items())[0]
            d[key] = list_of_values
        if remove_prefix:
            header = [h.split(".")[-1] for h in full_header]
        else:
            header = full_header
        num_observations = len(list(self[0].values())[0])
        rows = []
        # rows.append(header)
        for i in range(num_observations):
            row = [d[h][i] for h in full_header]
            rows.append(row)
        return header, rows

    def print_long(self):
        """ """
        for result in self:
            if hasattr(result, "combined_dict"):
                d = result.combined_dict
            else:
                d = result
            print_dict_with_rich(d)

    @convert_decorator
    def print(
        self,
        pretty_labels=None,
        filename=None,
        html=False,
        interactive=False,
        split_at_dot=True,
    ):
        if pretty_labels is None:
            pretty_labels = {}

        new_data = []
        for entry in self:
            key, list_of_values = list(entry.items())[0]
            new_data.append({pretty_labels.get(key, key): list_of_values})
        else:
            if not html:
                print_list_of_dicts_with_rich(
                    new_data, filename=filename, split_at_dot=split_at_dot
                )
            else:
                print_list_of_dicts_as_html_table(
                    new_data, filename=None, interactive=interactive
                )

    @convert_decorator
    def to_csv(self, filename: str = None, remove_prefix=False, download_link=False):
        """
        >>> r = create_example_results()
        >>> r.select('how_feeling').to_csv()
        'result.how_feeling\\r\\nBad\\r\\nBad\\r\\nGreat\\r\\nGreat\\r\\n'
        """
        header, rows = self._make_tabular(remove_prefix)

        if filename is not None:
            with open(filename, "w") as f:
                writer = csv.writer(f)
                writer.writerow(header)
                writer.writerows(rows)
        else:
            output = io.StringIO()
            writer = csv.writer(output)
            writer.writerow(header)
            writer.writerows(rows)

            if download_link:
                csv_file = output.getvalue()
                b64 = base64.b64encode(csv_file.encode()).decode()
                download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
                display(HTML(download_link))
            else:
                return output.getvalue()

    @convert_decorator
    def to_pandas(self, remove_prefix=False):
        csv_string = self.to_csv(remove_prefix=remove_prefix)
        csv_buffer = io.StringIO(csv_string)
        df = pd.read_csv(csv_buffer)
        return df

    @convert_decorator
    def to_dicts(self, remove_prefix=False):
        df = self.to_pandas(remove_prefix=remove_prefix)
        df = df.convert_dtypes()
        list_of_dicts = df.to_dict(orient="records")
        # Convert any pd.NA values to None
        list_of_dicts = [
            {k: (None if pd.isna(v) else v) for k, v in record.items()}
            for record in list_of_dicts
        ]
        return list_of_dicts

    @convert_decorator
    def to_list(self):
        if len(self) == 1:
            return list(self[0].values())[0]
        else:
            return tuple([list(x.values())[0] for x in self])


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from edsl.language_models.schemas import model_prices
from edsl.language_models.LanguageModel import LanguageModel
from edsl.language_models.model_interfaces.LanguageModelOpenAIThreeFiveTurbo import (
    LanguageModelOpenAIThreeFiveTurbo,
)
from edsl.language_models.model_interfaces.LanguageModelOpenAIFour import (
    LanguageModelOpenAIFour,
)
from edsl.language_models.model_interfaces.GeminiPro import GeminiPro



import json
import threading
import time
import uuid
from edsl.data import CRUD
from edsl.jobs.JobsRunner import JobsRunner
from edsl.results import Results


class JobsRunnerStreaming(JobsRunner):
    """This JobRunner conducts interviews serially."""

    runner_name = "streaming"

    def run(
        self,
        debug: bool = False,
        sleep: int = 0,
        n: int = 1,
        verbose: bool = False,
        progress_bar: bool = False,
    ) -> Results:
        """
        Conducts Interviews **serially** and returns their results.
        - `n`: how many times to run each interview
        - `debug`: prints debug messages
        - `verbose`: prints messages
        - `progress_bar`: shows a progress bar
        """
        job_uuid = str(uuid.uuid4())
        total_results = len(self.interviews)

        def conduct_and_save_interviews():
            for i, interview in enumerate(self.interviews):
                answer = interview.conduct_interview(debug=debug)
                CRUD.write_result(
                    job_uuid=job_uuid,
                    result_uuid=str(i),
                    agent=json.dumps(interview.agent.to_dict()),
                    scenario=json.dumps(interview.scenario.to_dict()),
                    model=json.dumps(interview.model.to_dict()),
                    answer=json.dumps(answer),
                )
                time.sleep(sleep)

        interview_thread = threading.Thread(target=conduct_and_save_interviews)
        interview_thread.start()

        return Results(
            survey=self.jobs.survey,
            data=[],
            job_uuid=job_uuid,
            total_results=total_results,
        )



from edsl.language_models.OpenAI import create_openai_model
from edsl.enums import LanguageModelType

model_name = LanguageModelType.GPT_4.value

LanguageModelOpenAIFour = create_openai_model(
    model_name=model_name, model_class_name="LanguageModelOpenAIFour"
)



from edsl.questions import Question, QuestionFunctional
from edsl.agents import Agent
from edsl.scenarios.Scenario import Scenario


def compose_questions(
    q1: Question, q2: Question, question_name: str = None
) -> QuestionFunctional:
    """
    Composes two questions. The answer to q1 is used as the input to q2.
    The resulting question is a question that can be used like other questions.
    Note:
    - One way the same result can be achieved is through q1+q2.
      Please see the __add__ method in Question.
    """
    if question_name is None:
        question_name = f"{q1.question_name}_{q2.question_name}"
    if q1.question_name not in q2.question_text:
        raise ValueError(
            f"q2 requires a field not present in q1's answer. "
            f"q1: {q1.question_name}, q2: {q2.question_name}"
        )

    def combo(
        scenario: Scenario, agent_traits: dict[str, str] = None
    ) -> QuestionFunctional:
        # get the answer to the first question
        first_answer = (
            q1.by(scenario)
            .by(Agent(traits=agent_traits))
            .run()
            .select(q1.question_name)[0]
        )
        # update the scenario with the answer to the first question
        scenario.update({q1.question_name: first_answer})
        # get the answer to the second question
        second_answer = (
            q2.by(scenario)
            .by(Agent(traits=agent_traits))
            .run()
            .select(q2.question_name)[0]
        )
        return second_answer

    return QuestionFunctional(
        question_name=question_name, question_text="functional", func=combo
    )


# UNCOMMENT BELOW TO SEE HOW THIS WORKS

# if __name__ == "__main__":
#     from edsl.questions import QuestionFreeText, QuestionFunctional
#     from edsl.questions.compose_questions import compose_questions
#     from edsl.scenarios.Scenario import Scenario

#     q1 = QuestionFreeText(
#         question_text="What is the capital of {{country}}", question_name="capital"
#     )
#     q2 = QuestionFreeText(
#         question_text="What is the population of {{capital}}",
#         question_name="population",
#     )
#     q3 = compose_questions(q1, q2)

#     jobs = q3.by(
#         Scenario({"country": "France"}),
#         Scenario({"country": "Germany"}),
#         Scenario({"country": "Greece"}),
#     )

#     print("Without an agent persona")
#     results1 = jobs.run()

#     print("Adding an agent persona")
#     results2 = jobs.by(
#         Agent(traits={"name": "Bob, who always mentions his travel agency."})
#     ).run()
#     results2.select("capital_population").table()

#     q1 = QuestionFreeText(
#         question_text="What is the capital of {{country}}", question_name="capital"
#     )
#     q2 = QuestionFreeText(
#         question_text="Is this {{population}} large?", question_name="population"
#     )
#     q3 = compose_questions(q1, q2)
#     print("Should throw an exception:")
#     try:
#         q3.by(Scenario({"country": "France"})).run()
#     except ValueError as e:
#         print(e)



import textwrap
from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class CheckBox(QuestionInstuctionsBase):
    question_type = "checkbox"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}
        Return a valid JSON formatted like this, selecting only the number of the option:
        {"answer": [<put comma-separated list of answer codes here>], "comment": "<put explanation here>"}
        {% if min_selections != None and max_selections != None and min_selections == max_selections %}
        You must select exactly {{min_selections}} options.
        {% elif min_selections != None and max_selections != None %}
        Minimum number of options that must be selected: {{min_selections}}.
        Maximum number of options that must be selected: {{max_selections}}.
        {% elif min_selections != None %}
        Minimum number of options that must be selected: {{min_selections}}.
        {% elif max_selections != None %}
        Maximum number of options that must be selected: {{max_selections}}.
        {% endif %}
        """
    )


class TopK(CheckBox):
    question_type = "top_k"



import uuid
from edsl.api.results import ResultsAPI


def JobRunnerAPI(api_key: str, job_dict: dict) -> ResultsAPI:
    api_key = api_key
    job_dict = job_dict
    # TODO: API call to send job data and return job_id
    job_id = str(uuid.uuid4())
    # return a ResultsAPI instance with the job_id
    return ResultsAPI(api_key, job_id)



from edsl.utilities.interface import (
    print_dict_as_html_table,
    print_dict_with_rich,
    print_list_of_dicts_as_html_table,
    print_list_of_dicts_with_rich,
    print_table_with_rich,
    print_public_methods_with_doc,
)

from edsl.utilities.utilities import (
    create_valid_var_name,
    dict_to_html,
    hash_value,
    HTMLSnippet,
    is_notebook,
    is_gzipped,
    is_valid_variable_name,
    random_string,
    repair_json,
    shorten_string,
)



from typing import Dict
from edsl.utilities.utilities import is_valid_variable_name


class TraitsDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """"""
        return instance.__dict__[self.name]

    def __set__(self, instance, traits_dict: Dict[str, str]) -> None:
        for key, value in traits_dict.items():
            if not is_valid_variable_name(key):
                raise Exception("Trait keys must be a valid variable name!")
        instance.__dict__[self.name] = traits_dict

    def __set_name__(self, owner, name: str) -> None:
        self.name = "_" + name


class CodebookDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """"""
        return instance.__dict__[self.name]

    def __set__(self, instance, codebook_dict: Dict[str, str]) -> None:
        instance.__dict__[self.name] = codebook_dict

    def __set_name__(self, owner, name: str) -> None:
        self.name = "_" + name


class InstructionDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """"""
        return instance.__dict__[self.name]

    def __set__(self, instance, instruction) -> None:
        instance.__dict__[self.name] = instruction
        instance.set_instructions = instruction != instance.default_instruction

    def __set_name__(self, owner, name: str) -> None:
        self.name = "_" + name



from __future__ import annotations
import textwrap
from random import uniform
from typing import Any, Optional, Union
from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions import Question
from edsl.questions.descriptors import NumericalOrNoneDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionNumerical(Question):
    """
    This question asks the user to answer with a numerical value.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `min_value` is the minimum value of the answer (float)
    - `max_value` is the maximum value of the answer (float)
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionNumerical.default_instructions`


    For an example, run `QuestionNumerical.example()`
    """

    question_type = "numerical"
    min_value: Optional[float] = NumericalOrNoneDescriptor()
    max_value: Optional[float] = NumericalOrNoneDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        min_value: Optional[Union[int, float]] = None,
        max_value: Optional[Union[int, float]] = None,
    ):
        self.question_name = question_name
        self.question_text = question_text
        self.min_value = min_value
        self.max_value = max_value

    ################
    # Answer methods
    ################
    def validate_answer(
        self, answer: dict[str, Any]
    ) -> dict[str, Union[str, float, int]]:
        """Validates the answer"""
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value_numeric(answer, "answer")
        self.validate_answer_numerical(answer)
        return answer

    def translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """There is no answer code."""
        return answer

    def simulate_answer(self, human_readable: bool = True):
        """Simulates a valid answer for debugging purposes"""
        return {
            "answer": uniform(self.min_value, self.max_value),
            "comment": random_string(),
        }

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionNumerical:
        return cls(
            question_name="age",
            question_text="How old are you in years?",
            min_value=0,
            max_value=86.7,
        )


def main():
    from edsl.questions.QuestionNumerical import QuestionNumerical

    q = QuestionNumerical.example()
    q.question_text
    q.min_value
    q.max_value
    # validate an answer
    q.validate_answer({"answer": 1, "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer(1)
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from __future__ import annotations
import random
import textwrap
from typing import Any, Optional, Union
from edsl.questions import Question
from edsl.questions.descriptors import (
    AllowNonresponseDescriptor,
    IntegerOrNoneDescriptor,
)
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionList(Question):
    """
    This question asks the user to answer by providing a list of items as comma-separated strings.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `max_list_items` is the maximum number of items that can be in the answer list (positive integer)
    - `allow_nonresponse` is whether the user can skip the question (boolean). If not provided, the default is False.
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionList.default_instructions`

    For an example, run `QuestionList.example()`
    """

    question_type = "list"
    max_list_items: int = IntegerOrNoneDescriptor()
    allow_nonresponse: bool = AllowNonresponseDescriptor()

    def __init__(
        self,
        question_text: str,
        question_name: str,
        allow_nonresponse: Optional[bool] = None,
        max_list_items: Optional[int] = None,
    ):
        self.question_text = question_text
        self.question_name = question_name
        self.allow_nonresponse = allow_nonresponse or False
        self.max_list_items = max_list_items

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: Any) -> dict[str, Union[list[str], str]]:
        """Validates the answer"""
        self.validate_answer_template_basic(answer)
        self.validate_answer_key_value(answer, "answer", list)
        self.validate_answer_list(answer)
        return answer

    def translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """There is no answer code."""
        return answer

    def simulate_answer(self, human_readable: bool = True):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        num_items = random.randint(1, self.max_list_items or 2)
        return {"answer": [random_string() for _ in range(num_items)]}

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionList:
        return cls(
            question_name="list_of_foods",
            question_text="What are your favorite foods?",
            allow_nonresponse=False,
            max_list_items=5,
        )


def main():
    from edsl.questions.QuestionList import QuestionList

    q = QuestionList.example()
    q.question_text
    q.question_name
    q.allow_nonresponse
    q.max_list_items
    # validate an answer
    q.validate_answer({"answer": ["pasta", "garlic", "oil", "parmesan"]})
    # translate answer code
    q.translate_answer_code_to_answer(["pasta", "garlic", "oil", "parmesan"])
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



from __future__ import annotations
import textwrap
import io
from abc import ABC, abstractmethod, ABCMeta
from rich.console import Console
from rich.table import Table
from jinja2 import Template, Environment, meta
from typing import Any, Type
from edsl.exceptions import (
    QuestionResponseValidationError,
    QuestionSerializationError,
    QuestionScenarioRenderError,
)
from edsl.questions.descriptors import (
    InstructionsDescriptor,
    QuestionNameDescriptor,
    QuestionTextDescriptor,
    ShortNamesDictDescriptor,
)

from edsl.prompts.Prompt import Prompt

from edsl.enums import QuestionType

# from edsl.questions.question_registry import get_question_class
from edsl.questions.AnswerValidatorMixin import AnswerValidatorMixin

from edsl.exceptions.questions import QuestionMissingTypeError, QuestionBadTypeError


class RegisterQuestionsMeta(ABCMeta):
    "Metaclass to register output elements in a registry i.e., those that have a parent"
    _registry = {}  # Initialize the registry as a dictionary

    def __init__(cls, name, bases, dct):
        super(RegisterQuestionsMeta, cls).__init__(name, bases, dct)
        if name != "Question":
            ## Enforce that all questions have a question_type class attribute
            ## and it comes from our enum of valid question types.
            if not hasattr(cls, "question_type"):
                raise QuestionMissingTypeError(
                    "Question must have a question_type class attribute"
                )

            if not QuestionType.is_value_valid(cls.question_type):
                acceptable_values = [item.value for item in QuestionType]
                raise QuestionBadTypeError(
                    f"""question_type must be one of {QuestionType} values, which are 
                                currently {acceptable_values}"""
                    ""
                )

            RegisterQuestionsMeta._registry[name] = cls

    @classmethod
    def get_registered_classes(cls):
        return cls._registry

    @classmethod
    def question_types_to_classes(
        cls,
    ):
        d = {}
        for classname, cls in cls._registry.items():
            if hasattr(cls, "question_type"):
                d[cls.question_type] = cls
            else:
                raise Exception(
                    f"Class {classname} does not have a question_type class attribute"
                )
        return d


from edsl.Base import PersistenceMixin, RichPrintingMixin


class Question(
    PersistenceMixin,
    RichPrintingMixin,
    ABC,
    AnswerValidatorMixin,
    metaclass=RegisterQuestionsMeta,
):
    """
    ABC for the Question class. All questions should inherit from this class.
    """

    question_name: str = QuestionNameDescriptor()
    question_text: str = QuestionTextDescriptor()
    short_names_dict: dict[str, str] = ShortNamesDictDescriptor()

    @property
    def data(self) -> dict:
        """Returns a dictionary of question attributes **except** for question_type"""
        candidate_data = {
            k.replace("_", "", 1): v
            for k, v in self.__dict__.items()
            if k.startswith("_")
        }
        optional_attributes = {
            "set_instructions": "instructions",
        }
        for boolean_flag, attribute in optional_attributes.items():
            if hasattr(self, boolean_flag) and not getattr(self, boolean_flag):
                candidate_data.pop(attribute, None)

        return candidate_data

    ############################
    # Serialization methods
    ############################
    def to_dict(self) -> dict[str, Any]:
        """Converts the question to a dictionary that includes the question type (used in deserialization)."""
        candidate_data = self.data.copy()
        candidate_data["question_type"] = self.question_type
        return candidate_data

    @classmethod
    def from_dict(cls, data: dict) -> Type[Question]:
        """Constructs a question object from a dictionary created by that question's `to_dict` method."""
        local_data = data.copy()
        try:
            question_type = local_data.pop("question_type")
        except:
            raise QuestionSerializationError(
                f"Data does not have a 'question_type' field (got {data})."
            )
        from edsl.questions.question_registry import get_question_class

        try:
            question_class = get_question_class(question_type)
        except ValueError:
            raise QuestionSerializationError(
                f"No question registered with question_type {question_type}"
            )
        return question_class(**local_data)

    ############################
    # Dunder methods
    ############################
    def __repr__(self) -> str:
        """Returns a string representation of the question. Should be able to be used to reconstruct the question."""
        class_name = self.__class__.__name__
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "question_type"
        ]
        return f"{class_name}({', '.join(items)})"

    def __eq__(self, other: Type[Question]) -> bool:
        """Checks if two questions are equal. Equality is defined as having the .to_dict()"""
        if not isinstance(other, Question):
            return False
        return self.to_dict() == other.to_dict()

    # TODO: Throws an error that should be addressed at QuestionFunctional
    def __add__(self, other_question):
        """
        Composes two questions into a single question.
        >>> from edsl.scenarios.Scenario import Scenario
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.questions.QuestionNumerical import QuestionNumerical
        >>> q1 = QuestionFreeText(question_text = "What is the capital of {{country}}", question_name = "capital")
        >>> q2 = QuestionNumerical(question_text = "What is the population of {{capital}}, in millions. Please round", question_name = "population")
        >>> q3 = q1 + q2
        """
        from edsl.questions import compose_questions

        return compose_questions(self, other_question)

    @abstractmethod
    def validate_answer(self, answer: dict[str, str]):
        pass

    def validate_response(self, response):
        """Validates the response from the LLM"""
        if "answer" not in response:
            raise QuestionResponseValidationError(
                "Response from LLM does not have an answer"
            )
        return response

    @abstractmethod
    def translate_answer_code_to_answer(self):  # pragma: no cover
        """Translates the answer code to the actual answer. Behavior depends on the question type."""
        pass

    @abstractmethod
    def simulate_answer(self, human_readable=True) -> dict:  # pragma: no cover
        """Simulates a valid answer for debugging purposes (what the validator expects)"""
        pass

    ############################
    # Forward methods
    ############################
    def add_question(self, other):
        "Adds a question to this question by turning them into a survey with two questions"
        from edsl.surveys.Survey import Survey

        s = Survey([self, other])
        return s

    def run(self, *args, **kwargs):
        "Turns a single question into a survey and runs it."
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s.run(*args, **kwargs)

    def by(self, *args):
        "Documentation missing."
        from edsl.surveys.Survey import Survey

        s = Survey([self])
        return s.by(*args)

    def rich_print(self):
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("Question Name", style="dim")
        table.add_column("Question Type")
        table.add_column("Question Text")
        table.add_column("Options")

        question = self
        if hasattr(question, "question_options"):
            options = ", ".join([str(o) for o in question.question_options])
        else:
            options = "None"
        table.add_row(
            question.question_name,
            question.question_type,
            question.question_text,
            options,
        )
        return table


if __name__ == "__main__":
    q = get_question_class("free_text")



from edsl.prompts.Prompt import PromptBase
from edsl.prompts.prompt_config import ComponentTypes


class QuestionInstuctionsBase(PromptBase):
    component_type = ComponentTypes.QUESTION_INSTRUCTIONS



from collections import UserDict


class Answers(UserDict):
    "Helper class to hold the answers to a survey"

    def add_answer(self, response, question) -> None:
        "Adds a response to the answers dictionary"
        answer = response.get("answer")
        comment = response.pop("comment", None)
        # record the answer
        self[question.question_name] = answer
        if comment:
            self[question.question_name + "_comment"] = comment

    def replace_missing_answers_with_none(self, survey) -> None:
        "Replaces missing answers with None. Answers can be missing if the agent skips a question."
        for question_name in survey.question_names:
            if question_name not in self:
                self[question_name] = None

    def to_dict(self):
        "Returns a dictionary of the answers"
        return self.data

    @classmethod
    def from_dict(cls, d):
        "Returns an Answers object from a dictionary"
        return cls(d)



import openai
import re
from typing import Any
from edsl import CONFIG
from openai import AsyncOpenAI
from edsl.enums import LanguageModelType, InferenceServiceType
from edsl.language_models import LanguageModel

LanguageModelType.GPT_4.value


def create_openai_model(model_name, model_class_name) -> LanguageModel:
    openai.api_key = CONFIG.get("OPENAI_API_KEY")

    class LLM(LanguageModel):
        """
        Child class of LanguageModel for interacting with OpenAI GPT-4 model.
        """

        _inference_service_ = InferenceServiceType.OPENAI.value
        _model_ = model_name
        _parameters_ = {
            "temperature": 0.5,
            "max_tokens": 1000,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "use_cache": True,
        }
        client = AsyncOpenAI()

        async def async_execute_model_call(
            self, user_prompt: str, system_prompt: str = ""
        ) -> dict[str, Any]:
            """Calls the OpenAI API and returns the API response."""
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
                temperature=self.temperature,
                max_tokens=self.max_tokens,
                top_p=self.top_p,
                frequency_penalty=self.frequency_penalty,
                presence_penalty=self.presence_penalty,
            )
            return response.model_dump()

        @staticmethod
        def parse_response(raw_response: dict[str, Any]) -> str:
            """Parses the API response and returns the response text."""
            response = raw_response["choices"][0]["message"]["content"]
            pattern = r"^```json(?:\\n|\n)(.+?)(?:\\n|\n)```$"
            match = re.match(pattern, response, re.DOTALL)
            if match:
                return match.group(1)
            else:
                return response

    LLM.__name__ = model_class_name

    return LLM


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import copy
from collections import UserDict
from rich.table import Table

from edsl.Base import Base


class Scenario(Base, UserDict):
    """A Scenario is a dictionary of key/values that describe some situation."""

    def __add__(self, other_scenario):
        """Combine two scenarios. If the other scenario is None, then just return self.
        >>> s1 = Scenario({"price": 100, "quantity": 2})
        >>> s2 = Scenario({"color": "red"})
        >>> s1 + s2
        {'price': 100, 'quantity': 2, 'color': 'red'}
        >>> (s1 + s2).__class__.__name__
        'Scenario'
        """
        if other_scenario is None:
            return self
        else:
            new_scenario = Scenario()
            new_scenario.data = copy.deepcopy(self.data)
            new_scenario.update(copy.deepcopy(other_scenario))
            return Scenario(new_scenario)

    def to(self, question_or_survey) -> "Jobs":
        """Run a question/survey with this particular scenario.
        Useful if you want to reverse the typical chain of operations.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> s = Scenario({"food": "wood chips"})
        >>> q = QuestionMultipleChoice(question_text = "Do you enjoy the taste of {{food}}?", question_options = ["Yes", "No"], question_name = "food_preference")
        >>> _ = s.to(q)
        """
        return question_or_survey.by(self)

    def rename(self, replacement_dict: dict) -> "Scenario":
        """Rename the keys of a scenario. Useful for changing the names of keys.

        >>> s = Scenario({"food": "wood chips"})
        >>> s.rename({"food": "food_preference"})
        {'food_preference': 'wood chips'}
        """
        new_scenario = Scenario()
        for key, value in self.items():
            if key in replacement_dict:
                new_scenario[replacement_dict[key]] = value
            else:
                new_scenario[key] = value
        return new_scenario

    def make_question(self, question_class: type):
        """Make a question from this scenario. Note it takes a QuestionClass (not a question)
        as an input.

        >>> from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
        >>> from edsl.agents.Agent import Agent

        >>> s = Scenario({"question_name": "feelings",
        ...               "question_text": "How are you feeling?",
        ...               "question_options": ["Very sad.", "Sad.", "Neutral.", "Happy.", "Very happy."]})
        >>> q = s.make_question(QuestionMultipleChoice)
        >>> q.by(Agent(traits = {'feeling': 'Very sad'})).run().select("feelings")
        [{'answer.feelings': ['Very sad.']}]
        """
        return question_class(**self)

    def to_dict(self):
        """Convert a scenario to a dictionary.
        >>> s = Scenario({"food": "wood chips"})
        >>> s.to_dict()
        {'food': 'wood chips'}
        """
        return self.data

    @classmethod
    def from_dict(cls, d):
        """Convert a dictionary to a scenario.
        >>> Scenario.from_dict({"food": "wood chips"})
        {'food': 'wood chips'}
        """
        return cls(d)

    def _table(self) -> tuple[dict, list]:
        """Prepare generic table data."""
        table_data = []
        for attr_name, attr_value in self.__dict__.items():
            table_data.append({"Attribute": attr_name, "Value": repr(attr_value)})
        column_names = ["Attribute", "Value"]
        return table_data, column_names

    def rich_print(self):
        """Displays an object as a rich table."""
        table_data, column_names = self._table()
        table = Table(title=f"{self.__class__.__name__} Attributes")
        for column in column_names:
            table.add_column(column, style="bold")

        for row in table_data:
            row_data = [row[column] for column in column_names]
            table.add_row(*row_data)

        return table

    @classmethod
    def example(cls):
        """Returns an example scenario.
        >>> Scenario.example()
        {'persona': 'A reseacher studying whether LLMs can be used to generate surveys.'}
        """
        return cls(
            {
                "persona": "A reseacher studying whether LLMs can be used to generate surveys."
            }
        )

    def code(self):
        """Returns the code for the scenario."""
        lines = []
        lines.append("from edsl.scenario import Scenario")
        return f"Scenario({self.data})"


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations
from edsl.questions.descriptors import QuestionOptionsDescriptor
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice


class QuestionYesNo(QuestionMultipleChoice):
    """
    This question asks the user to respond with "Yes" or "No".

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)
    - `question_options` are the options the user should select from (list of strings)

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionYesNo.default_instructions`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, see `QuestionYesNo.example()`
    """

    question_type = "yes_no"
    question_options = QuestionOptionsDescriptor(num_choices=2)

    def __init__(
        self,
        question_name: str,
        question_text: str,
        short_names_dict: dict[str, str] = None,
        question_options: list[str] = ["Yes", "No"],
    ):
        super().__init__(
            question_name=question_name,
            question_text=question_text,
            question_options=question_options,
            short_names_dict=short_names_dict,
        )
        self.question_options = question_options

    ################
    # Helpful
    ################
    @classmethod
    def example(cls) -> QuestionYesNo:
        return cls(
            question_name="is_it_raining",
            question_text="Is it raining?",
            short_names_dict={"Yes": "y", "No": "y"},
        )


def main():
    from edsl.questions.derived.QuestionYesNo import QuestionYesNo

    q = QuestionYesNo.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    # validate an answer
    q.validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer(0, {})
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



class ResultsAPI:
    """
    While results = None, repr will be calling the api to get the results or the estimated time to completion. When it does, results will be stored here, and the API will no longer be called.
    """

    def __init__(self, api_key: str, job_id: str) -> None:
        self.api_key = api_key
        self.job_id = job_id
        self.status = None
        self.results = None

    def _get_results(self) -> dict:
        # TODO: API call to get status & results
        return ("Running", None)

    def __repr__(self) -> str:
        if self.results is None:
            self.status, self.results = self._get_results()
        return f"Results(api_key={self.api_key}, job_id={self.job_id}, status={self.status}, results={self.results})"
